package com.tidecloud.dataacceptance.service.impl;

import com.accept.client.DeviceMsgClient;
import com.tidecloud.dataacceptance.codec.HeaderTailDelimiterFrameDecoder;
import com.tidecloud.dataacceptance.common.*;
import com.tidecloud.dataacceptance.entity.YiTongGPSDevice;
import com.tidecloud.dataacceptance.service.HexBinaryAcceptanceHandlerAdapter;
import com.tidecloud.dataacceptance.util.FileUtils;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.xml.bind.DatatypeConverter;
import java.util.Calendar;
import java.util.Date;

/**
 * @author jhw
 * GK309 卡牌数据 康凯斯
 */
@Component
@ChannelHandler.Sharable
@Scope("prototype")
public class GK309GpsServerHandler extends HexBinaryAcceptanceHandlerAdapter {

    /**
     * 协议信息数据
     * [ 起始位|包长度|协议号|信息内容|信息序列号|错误校验|停止位 ]
     */
    private static final Logger logger = LoggerFactory.getLogger(GK309GpsServerHandler.class);

    // 起始位
    private static final Integer START_BITS = 2;// 起始位长度
    private static final Byte START_BIT = 0x78; //单字节长度数据包
    private static final Byte START_BIT2 = 0X79;// 双字节长度数据包

//    private static final int START_BIT = 120; //单字节长度数据包
//    private static final int START_BIT2 = 121;// 双字节长度数据包

    // 协议号
    private static final byte LOGIN_MSG = 0x01;// 登录 (必须回复)
    private static final byte STATUS_MSG = 0x13;// 心跳，状态信息包  (必须回复)
    private static final byte LOCATION_MSG = 0x10;// 一般GPS 位置信息上传
    private static final byte GPS_WARN = 0x16;// 一般GPS 位置信息上传
    private static final byte LBS_WARN = 0x19;// 一般GPS 位置信息上传
    private static final byte CORRECT_TIME_MSG = 0x1F;// 校验时间
    private static final byte SIGN_IN_MSG = (byte)0xB0;// 打卡

    private static final byte[] END_BITS = new byte[]{0X0D, 0x0A};

    private static final Integer DATA_SIZE = 6;

    private static final byte MARK = 0B00111100;

    private static final byte SIGN_IN = 0B00000100; //上班打卡
    private static final byte SIGN_OUT = 0B00001000; // 下班打卡

    @Autowired
    private DeviceMsgClient deviceMsgClient;

    /**
     * <b>协议格式：</b>
     *  <p>起始位 2位
     *  <p>包长度 1位 :协议号+信息内容+信息序列号+错误校验
     *  <p>协议号 1位
     *  <p>信息内容 看每个协议的规定
     *  <p>信息序列号 2位
     *  <p>错误校验 2位
     *  <p>停止位 2位
     *  <p></p>
     * <p>本方法把起始位去掉，拿出包长度和剩下的部分调用{@link #handle(ByteBuf, int, Channel)} 方法
     * @param in
     * @param channel
     * @throws Exception
     */
    @Override
    protected void handle(ByteBuf in, Channel channel) throws Exception {
        if (in.isReadable()) {
            in.markReaderIndex();
            try {
                byte b = 0;
                int index = 0;
                while (index < START_BITS) {
                    b = in.readByte();
                    if (START_BIT2 != b && START_BIT != b) {
                        channel.close();
                    }
                    index++;
                }
                /*  单字节长度或者双字节长度 */
                int length = 0;
                if (START_BIT == b) {
                    length = in.readByte() & 0xff; // 一个字节
                } else {
                    length = in.readShort() & 0xffff;// 两个字节
                }
                handle(in, length, channel);
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
        }
    }

    /**
     * 根据协议号处理每个指令
     * @param in 协议号及其后面的报文组成的byteBuf
     * @param length 协议长度
     * @param channel
     * @throws Exception
     */
    private void handle(ByteBuf in, int length, Channel channel) throws Exception {
        if (in.isReadable()) {
            byte msgType = in.readByte();// 协议号 1个字节
            String deviceId = channelDeviceMap.get(channel);
            if (deviceId != null) {
                MDC.put(MDC_DEVICEID, deviceId);
            } else {
                if (LOGIN_MSG != msgType) {
                    channel.close();
                }
            }
            if (LOGIN_MSG == msgType) {
                resolveLoginMSG(in, channel);
            } else if (STATUS_MSG == msgType) {
                reply(channel, STATUS_MSG, "心跳包");
                byte[] bytes = byteMerger(getOriginalData(in, deviceId), DateUtil.formatDate2String(new Date()).getBytes());
                sendMsg2Kafka(bytes, deviceId, channel);// 心跳数据对接
                deviceMsgClient.acceptDeviceMsgParam("kangkaisi",
                        deviceId,1,
                        DatatypeConverter.printHexBinary(bytes), System.currentTimeMillis());
            } else if (LOCATION_MSG == msgType) {
                logger.info("GPS 定位包");
                // resolveLocationMSG(in,deviceId);
                byte[] bytes = byteMerger(getOriginalData(in, deviceId), DateUtil.formatDate2String(new Date()).getBytes());
                sendMsg2Kafka(bytes, deviceId, channel);
                deviceMsgClient.acceptDeviceMsgParam("kangkaisi",
                        deviceId,2,
                        DatatypeConverter.printHexBinary(bytes), System.currentTimeMillis());
            } else if (CORRECT_TIME_MSG == msgType) {
                // 1位长度位和6位时间位,后面才是序列号
                in.skipBytes(7);
                short x = in.readShort();
                byte[] ret = short2Bytes(x);

//                reply(channel, CORRECT_TIME_MSG, "校时包");
                replyForTime(channel, ret);
            } else if (SIGN_IN_MSG == msgType) {
                replayForSignIn(in, length, channel);
            } else if(GPS_WARN == msgType) {
                replyGpsWarn(channel);
            } else if (LBS_WARN == msgType) {
                replyLbsWarn(channel);
            } else {
                logger.info("client send data without handle type ...");
            }
        }

    }

    /**
     * 对于报警消息，设备端能够拿到经纬度信息，但是拿不到地址信息，所以需要服务端根据消息体中的经纬度回复对应的
     * 地址信息。在太的的使用场景中，不会使用到这个地址，<b>所以目前的做法是回复一个固定的地址,，只有序列号是变化的。</b>
     * <p>
     * <p>按照一下示例进行回复的:
     * <p>设备上传：78 78 25 16 14 03 0B 0F 12 12 CC 02 6C 19 06 0C 38 D1 92 00 15 45 09 01 CC 00 28 66 00 0F 6F 20 03 04 00
     * 01 00 0C 8C CA 0D 0A
     * <p>平台回复：78 78 76 16 70 36 4E 55 30 41 44 44 52 45 53 53 26 26 00 53 00 4F 00 53 62 A5 8B 66 00 3A 5E 7F 4E 1C 77
     * 01 6D F1 57 33 5E 02 5B 9D 5B 89 53 3A 96 86 66 0C 8D EF 00 32 00 2C 9A D8 65 B0 59 47 79 D1 62 80 56 ED 4E 1C
     * 53 57 00 33 00 34 7C 73 FF 08 7C BE 78 6E 5B 9A 4F 4D FF 09 26 26 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
     * 00 00 00 00 00 00 23 23 00 0C 23 04 0D 0A
     */
    private void replyGpsWarn(Channel channel) {
        byte[] contentToReply = new byte[]{0x76, 0x16, 0x70, 0x36, 0x4E, 0x55, 0x30, 0x41, 0x44, 0x44,
                0x52, 0x45, 0x53, 0x53, 0x26, 0x26, 0x00, 0x53, 0x00, 0x4F, 0x00, 0x53, 0x62, (byte) 0xA5,
                (byte) 0x8B, 0x66, 0x00,
                0x3A, 0x5E, 0x7F, 0x4E, 0x1C, 0x77, 0x01, 0x6D, (byte) 0xF1, 0x57, 0x33, 0x5E, 0x02, 0x5B,
                (byte) 0x9D, 0x5B, (byte) 0x89,
                0x53, 0x3A, (byte) 0x96, (byte) 0x86, 0x66, 0x0C, (byte) 0x8D, (byte) 0xEF, 0x00, 0x32, 0x00, 0x2C,
                (byte) 0x9A, (byte) 0xD8, 0x65, (byte) 0xB0,
                0x59,
                0x47, 0x79, (byte) 0xD1, 0x62, (byte) 0x80, 0x56, (byte) 0xED, 0x4E, 0x1C, 0x53, 0x57, 0x00, 0x33,
                0x00, 0x34, 0x7C, 0x73,
                (byte) 0xFF, 0x08, 0x7C, (byte) 0xBE, 0x78, 0x6E, 0x5B, (byte) 0x9A, 0x4F, 0x4D, (byte) 0xFF, 0x09,
                0x26, 0x26, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x23, 0x23, 0x00, 0x0C};
        writeToClient(channel, "GPS报警报", contentToReply);
    }

    /**
     * 对于报警消息，设备端能够拿到经纬度信息，但是拿不到地址信息，所以需要服务端根据消息体中的经纬度回复对应的
     * 地址信息。在太的的使用场景中，不会使用到这个地址，<b>所以目前的做法是回复一个固定的地址,，只有序列号是变化的。</b>
     * <p>
     * <p>按照一下示例进行回复的:
     * <p>LBS.状态包（终端?服务器)：
     * <p>78 78 12 19 01 CC 00 28 66 00 0E EE 20 05 04 00 01 00 14 50 0C 0D 0A
     * <p>服务器回复（服务器?终端）：
     * <p>紧 急 呼 叫 ： 广 东 省 . 深 圳 市 . 宝 安 区 边 检 路 . 甲 岸 科 技 园 附 近 . 同 乐 出 口 附
     * <p>近.(N22.575,E113.917)附近
     * <p>78 78 9A 17 94 00 00 00 01 41 44 44 52 45 53 53 26 26 7D 27 60 25 54 7C 53 EB 00
     * <p>3A 5E 7F 4E 1C 77 01 00 2E 6D F1 57 33 5E 02 00 2E 5B 9D 5B 89 53 3A 8F B9 68 C0
     * <p>8D EF 00 2E 75 32 5C B8 79 D1 62 80 56 ED 96 44 8F D1 00 2E 54 0C 4E 50 51 FA 53
     * <p>E3 96 44 8F D1 00 2E 00 28 00 4E 00 32 00 32 00 2E 00 35 00 37 00 35 00 2C 00 45
     * <p>00 31 00 31 00 33 00 2E 00 39 00 31 00 37 00 29 96 44 8F D1 26 26 00 00 00 00 00
     * <p>00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 23 23 00 14 87 07 0D 0A
     */
    private void replyLbsWarn(Channel channel) {
        byte[] contentToReply = new byte[]{
                (byte) 0x9A, 0x17, (byte) 0x94, 0x00, 0x00, 0x00, 0x01, 0x41, 0x44, 0x44, 0x52, 0x45, 0x53, 0x53, 0x26,
                0x26, 0x7D, 0x27, 0x60, 0x25, 0x54, 0x7C, 0x53, (byte) 0xEB, 0x00,
                0x3A, 0x5E, 0x7F, 0x4E, 0x1C, 0x77, 0x01, 0x00, 0x2E, 0x6D, (byte) 0xF1, 0x57, 0x33, 0x5E, 0x02, 0x00
                , 0x2E,
                0x5B, (byte) 0x9D, 0x5B, (byte) 0x89, 0x53, 0x3A, (byte) 0x8F, (byte) 0xB9, 0x68, (byte) 0xC0,
                (byte) 0x8D, (byte) 0xEF, 0x00, 0x2E, 0x75, 0x32, 0x5C, (byte) 0xB8, 0x79, (byte) 0xD1, 0x62,
                (byte) 0x80, 0x56, (byte) 0xED, (byte) 0x96, 0x44, (byte) 0x8F,
                (byte) 0xD1, 0x00, 0x2E, 0x54, 0x0C, 0x4E, 0x50, 0x51, (byte) 0xFA, 0x53,
                (byte) 0xE3, (byte) 0x96, 0x44, (byte) 0x8F, (byte) 0xD1, 0x00, 0x2E, 0x00, 0x28, 0x00, 0x4E, 0x00,
                0x32, 0x00, 0x32, 0x00, 0x2E,
                0x00, 0x35, 0x00, 0x37, 0x00, 0x35, 0x00, 0x2C, 0x00, 0x45,
                0x00, 0x31, 0x00, 0x31, 0x00, 0x33, 0x00, 0x2E, 0x00, 0x39, 0x00, 0x31, 0x00, 0x37, 0x00, 0x29,
                (byte) 0x96,
                0x44, (byte) 0x8F, (byte) 0xD1, 0x26, 0x26, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x23,
                0x23, 0x00, 0x14
        };
        writeToClient(channel, "LBS报警报", contentToReply);
    }

    private void replayForSignIn(ByteBuf in, int length, Channel channel) {
        byte[] dataBytes = new byte[6];
        in.readBytes(dataBytes);

        in.skipBytes(1);
        byte[] reservedWorkBytes = new byte[2];
        in.readBytes(reservedWorkBytes);

        // 根据文档查看上下班标示
        ByteBuf byteBuf = null;
        byte b;
        try {
            byteBuf = in.readBytes(13);
            byteBuf.skipBytes(12);
            b = byteBuf.readByte();
        } finally {
            if (byteBuf != null) {
                byteBuf.release();
            }
        }
        // 2-5位代表是上班还是下班
        byte i = (byte) (b & MARK);

//                类型: 1(上班), 2(下班)
        Byte signResponse = null;
        if (i == SIGN_IN) {
            signResponse = 0X01;
        } else {
            signResponse = 0x02;
        }

        // 减去前面跳过的第一位和前面跳过的7位，再减去最后的四位（序列号本身的2位+校验位2位）
        in.skipBytes(length - 8 - 2 - 2 - 15);
        short serNum = in.readShort();
        byte[] serNumBytes = short2Bytes(serNum);

        // 状态： 0(失败), 1(成功)
        byte[] signResult = new byte[]{0x01, signResponse};
        // 0X0F是长度，固定长度
        byte[] contentsOfPackage = mergeByteArray(new byte[]{0X0F, SIGN_IN_MSG},
                dataBytes, signResult, reservedWorkBytes, serNumBytes);
        writeToClient(channel, "打卡包", contentsOfPackage);
    }

    private byte[] short2Bytes(short x) {
        byte[]  ret = new byte[2];
        ret[0] = (byte)(x & 0xff);
        ret[1] = (byte)((x >> 8) & 0xff);
        return ret;
    }

    /**
     * 登录管理
     *
     * @param in
     * @param channel
     */
    private void resolveLoginMSG(ByteBuf in, Channel channel) {
        /**  终端ID|类型识别码|扩展位*/
        byte[] deviceIdBytes = new byte[8];// 终端ID [8位]
        in.readBytes(deviceIdBytes);
        String deviceId = DatatypeConverter.printHexBinary(deviceIdBytes);
        String deviceIdInMap = channelDeviceMap.get(channel);
        MDC.put(MDC_DEVICEID, deviceId);
        if (!deviceId.equals(deviceIdInMap)) {
            manageChannel(channel, deviceId);
        }
        reply(channel, LOGIN_MSG, "登录包");
    }

    public static byte[] byteMerger(byte[] byte_1, byte[] byte_2) {
        byte[] byte_3 = new byte[byte_1.length + byte_2.length];
        System.arraycopy(byte_1, 0, byte_3, 0, byte_1.length);
        System.arraycopy(byte_2, 0, byte_3, byte_1.length, byte_2.length);
        return byte_3;
    }


    /**
     * 回复包
     *
     * @param channel
     * @param msgType
     */
    private void reply(Channel channel, byte msgType, String msg) {

        /**
         * 起始位[]长度[] 协议号[]	序列号[]CRC校验[]停止位[2]
         */
        ByteBuf buffer = Unpooled.buffer();
        byte[] crcBytes = new byte[]{0x05, msgType, 0x00, 0x05}; // CRC校验数据
        int doCrc = CRCUtil.do_crc(65535, crcBytes);//CRC-ITU
        byte[] intToByte = NumUtil.intToByte(doCrc, 2);
        byte[] bytes = new byte[]{0x78, 0x78, 0x05, msgType, 0x00, 0x05, intToByte[0], intToByte[1], 0x0D, 0x0A};
        buffer.writeBytes(bytes);
        ChannelFuture channelFuture = channel.write(buffer);
        channelFuture.addListener(future -> {
            String deviceId = channelDeviceMap.get(channel);
            logger.info("server reply [{}] to client device [{}] success:[{}] ", msg, deviceId, DatatypeConverter.printHexBinary(bytes));
        });
    }

    private byte[] mergeByteArray(byte[]... bytesArray) {
        byte[] result = ArrayUtils.EMPTY_BYTE_ARRAY;
        for (byte[] bytes : bytesArray) {
            result = ArrayUtils.addAll(result, bytes);
        }
        return result;
    }

    /**
     * 校时回复
     * @param channel
     */
    private void replyForTime(Channel channel, byte[] serialNumber) {
        byte msgType = CORRECT_TIME_MSG;
        String msg = "校验时间";

        /**
         * 起始位[]长度[] 协议号[]	序列号[]CRC校验[]停止位[2]
         */
        long utcSeconds = Calendar.getInstance().getTimeInMillis() / 1000L;

        // 长度(固定0B)-协议号(固定）-utc时间秒数-预留位(2个0x00)-序列号
        byte[] utcBytes = BitOperator.longToBytes(utcSeconds, 4);
        byte[] seriNumbytes = mergeByteArray(new byte[]{0x00, 0x00}, serialNumber);
        byte[] contentsOfPackage = mergeByteArray(new byte[]{0x0B, msgType}, utcBytes, seriNumbytes);

        writeToClient(channel, msg, contentsOfPackage);
    }

    /**
     * 回复给客户端
     * @param channel
     * @param msg
     * @param contentsOfPackage
     */
    private void writeToClient(Channel channel, String msg, byte[] contentsOfPackage) {
        ByteBuf buffer = Unpooled.buffer();
        byte[] bytesToReply = wrapeContents(contentsOfPackage);
        buffer.writeBytes(bytesToReply);
        ChannelFuture channelFuture = channel.write(buffer);
        channelFuture.addListener(future -> {
            String deviceId = channelDeviceMap.get(channel);
            logger.info("server reply [{}] to client device [{}] success:[{}] ", msg, deviceId, DatatypeConverter.printHexBinary(bytesToReply));
        });
    }

    /**
     * 对数据做CRC校验并添加校验位/开始位置/结束位置
     * @param contentsOfPackage 数据内容指的是去除起始位，校验位和结束位的部分: startBit + contentsOfPackage + crcBytes + endbit
     * @return 可以直接写给客户端的数据
     */
    private byte[] wrapeContents(byte[] contentsOfPackage) {
        int doCrc = CRCUtil.do_crc(65535, contentsOfPackage);//CRC-ITU
        byte[] crcBytes = NumUtil.intToByte(doCrc, 2);

        return mergeByteArray(new byte[]{START_BIT, START_BIT}, contentsOfPackage, crcBytes, END_BITS);
    }

    /**
     * 原始数据重新组装
     *
     * @param in
     * @param deviceId
     * @return
     */
    private byte[] getOriginalData(ByteBuf in, String deviceId) {
        if (StringUtils.isNotBlank(deviceId)) {
            in.resetReaderIndex();
            byte[] deviceArr = deviceId.getBytes();
            int length = in.readableBytes();
            byte[] dataByteArray = new byte[length + deviceArr.length];
            in.readBytes(dataByteArray, 0, in.readableBytes());
            System.arraycopy(deviceArr, 0, dataByteArray, length, deviceArr.length);
            return dataByteArray;
        } else
            return null;
    }

    private void resolveLocationMSG(ByteBuf in, String deviceId) throws Exception {
        YiTongGPSDevice yiTongGPSDevice = new YiTongGPSDevice();
        StringBuffer dateTimeStrBuf = new StringBuffer();
        int indexOfDateTime = 0;
        while (indexOfDateTime < DATA_SIZE) {
            byte b = in.readByte();
            dateTimeStrBuf.append(NumUtil.byte2String(b));
            indexOfDateTime++;
        }
        // 日期
        yiTongGPSDevice.setDate(dateTimeStrBuf.toString());
        // gps信息卫星数
        yiTongGPSDevice.setGpsCount(in.readByte());
        // 维度
        yiTongGPSDevice.setLat(in.readInt());
        // 经度
        yiTongGPSDevice.setLng(in.readInt());
        // 速度
        yiTongGPSDevice.setSpeedbyte(in.readByte());
        // 航向
        yiTongGPSDevice.setCourseStatus(in.readShort());
        // 国家代号
        yiTongGPSDevice.setMcc(in.readShort());
        // 移动网号码
        yiTongGPSDevice.setMnc(in.readByte());
        // 位置区码
        yiTongGPSDevice.setLac(in.readShort());
        // 移动基站Cell Tower ID
        yiTongGPSDevice.setCellId(in.readMedium());
        yiTongGPSDevice.setAcc(in.readByte());
        // 数据上报模式 0x00:定时上报,0x01:定距上报,0x02:拐点上传,0x03:ACC状态改变上传,0X08:开机上报位置信息
        yiTongGPSDevice.setReportModel(in.readByte());
        // 0x01:实时 0x00:补传
        yiTongGPSDevice.setIsmendMsg(in.readByte());
        double mileage = NumUtil.toFixed2Place((double) in.readInt() / 1000);
        // 里程设备默认是关闭的，需要指令，设备端才发送
        yiTongGPSDevice.setDeviceId(deviceId);
        yiTongGPSDevice.setMileage(mileage);
        yiTongGPSDevice.setDataType(1);
        // 写文件操作
        String deviceStr = YiTongGPSDevice.buildDeviceStr(yiTongGPSDevice);
        FileUtils.dataStorage(deviceStr.getBytes(), dataPath, prefixName);
    }

    /**
     * 按照开始，结束标记位读取数据
     */
    @Override
    public void startAcceptor() {
        EventLoopGroup bossGroup = new NioEventLoopGroup();
        EventLoopGroup workerGroup = new NioEventLoopGroup();
        byte[] tailSplitBytes = new byte[]{0x0D, 0x0A};
        byte[] headerSplitBytes = new byte[]{0x78, 0x78};
        byte[] headerSplitBytes1 = new byte[]{0x79, 0x79};
        try {
            ServerBootstrap b = new ServerBootstrap();
            b.group(bossGroup, workerGroup).channel(NioServerSocketChannel.class)
                    .childHandler(new ChannelInitializer<SocketChannel>() {
                        @Override
                        protected void initChannel(SocketChannel ch) throws Exception {
                            ch.pipeline().addLast(new HeaderTailDelimiterFrameDecoder(65535, false,
                                    Unpooled.copiedBuffer(tailSplitBytes), Unpooled.copiedBuffer(headerSplitBytes),
                                    Unpooled.copiedBuffer(headerSplitBytes1)));
                            ch.pipeline().addLast(GK309GpsServerHandler.this);
                        }
                    }).option(ChannelOption.SO_BACKLOG, 128).childOption(ChannelOption.SO_KEEPALIVE, true);
            ChannelFuture f = b.bind(port).sync();
            f.channel().closeFuture().sync();
            logger.info("start accept service for {}, bind address {}:{}", this.getPrefixName(), this.getIp(),
                    this.getPort());
        } catch (InterruptedException e) {
            logger.error(e.getMessage());
        } finally {
            cleanRedisLinkData();
            workerGroup.shutdownGracefully();
            bossGroup.shutdownGracefully();
        }
    }

}
